/*
 *  Copyright (C) 2000, Sealevel Systems
 *   
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  The author may be contacted via email at dalew@sealevel.com.  This is 
 *  NOT a technical support address; contact your vendor for help installing
 *  and configuring the software.
 */
#include "libseaio.h"
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include <string.h>

#define CLOSE_RELAY 1
#define OPEN_RELAY 0

extern int errno;

int SeaIo_WriteReg(int hDevice, unsigned char bRegNumber,
		   unsigned char Data)
{
	Seaio_Device			device			= {0};
	Seaio_WritePortStruct	writePortStruct;
	int success = 0, i=0;

	if ( ioctl(hDevice, SEAIO_IOCGDESCRIPTOR, &device) < 0 ) {
		return Seaio_Error(errno);
	}

	if (device.writable == 0)
		return SEAIO_ERROR_NOT_SUPPORTED;

	for (i = 0; (i < device.writable) && (device.output[i] != -1); i++) {
		if (device.output[i] == bRegNumber) {
			writePortStruct.portOffset	= bRegNumber;
			writePortStruct.dataByte	= Data;
			if(ioctl(hDevice, SEAIO_IOCWRITEPORT, &writePortStruct) < 0)
				return Seaio_Error(errno);
			success = 1;
			break;
		}
	}
	
	if (success != 1)
		return SEAIO_ERROR_PARAM;
	else
		return SEAIO_ERROR_NONE;
}

int SeaIo_ReadAnalog(int hDevice, unsigned long Channel,
			unsigned long * Data)
{
	Seaio_Device			device						= {0};
	UCHAR 						configByte, inByte;
	int								retCode, i;
	unsigned long 		RetLength;

	/* Get our device info */
	if ( ioctl(hDevice, SEAIO_IOCGDESCRIPTOR, &device) < 0 )
	{
		return Seaio_Error(errno);
	}

	/* Setup A/D control information */
	fprintf(stderr,"Current channel %ld range is: %x \n",Channel,device.ucAIRange[Channel]);
	configByte = 0x040;	// normal power / internal clock mode / internal acquisition
	configByte |= (device.ucAIRange[Channel] & 0x03) << 3;
	configByte |= (Channel & 0x07);

	/* Configure the channel and initiate A/D acquisition cycle */
  retCode =	SeaIo_WriteReg(hDevice, 0, configByte);
	if(retCode != SEAIO_ERROR_NONE)
	{
		return retCode;
	}

	/* Poll for acquisition cycle completion */
	/* Windows driver attempts for up to 500 microseconds */
	retCode = SEAIO_ERROR_INCOMPLETE;
	for(i=0; i<MAX_AD_RETRIES; i++)
	{
			/* Check for completion */
			SeaIo_ReadReg(hDevice, 2, &inByte, 1, &RetLength);
			if( !(inByte & 0x01) )
			{
				/* Acquisition cycle is complete */
				retCode = SEAIO_ERROR_NONE;
				break;
			}

			/* Delay 100 usecs before checking completion status again */
			usleep( 100 );

	}	

	/* Verify cycle completed successfully */
	if( retCode != SEAIO_ERROR_NONE )
	{
		return retCode;
	}

	/* Return A/D value from requested channel */
	/* Upper 4 bits */
	SeaIo_ReadReg(hDevice, 1, &inByte, 1, &RetLength);
	*Data = (inByte & 0x0f) << 8;

	/* Remaining 8 bits */
	SeaIo_ReadReg(hDevice, 0, &inByte, 1, &RetLength);
	*Data |= inByte;

	/* Success - we are done with this channel */
	return SEAIO_ERROR_NONE;
}

/* This function is currently only valid on a 3820 */
int SeaIo_WriteAnalog(int hDevice, unsigned long Channel,
			unsigned long Data)
{
	Seaio_Device			device						= {0};
	UCHAR 						hiByte, lowByte;
	int								retCode;

	/* Validate our parameters */
  if( Channel < 0 || Channel > 1)
	{
		fprintf(stderr,"Invalid channel \n");
		return SEAIO_ERROR_PARAM;
	}

	/* Get our device info */
	if ( ioctl(hDevice, SEAIO_IOCGDESCRIPTOR, &device) < 0 )
	{
		return Seaio_Error(errno);
	}

	/* Validate our card type */
	if( device.model != 0x3820 && device.model != 3820 )
	{	
		fprintf(stderr,"Not supported for this card: %x \n",device.model);
		return SEAIO_ERROR_NOT_SUPPORTED;
	}

	/* Extract register data from "Data" input */
	lowByte = (UCHAR)Data;
	hiByte = (UCHAR)(Data >> 8);

	/* Possible improvement here is to add channels and register info to
     the card type array */
	/* Write the data bytes to the correct channel and strobe the DAC */
	if( Channel == 0 )
	{
		retCode =	SeaIo_WriteReg(hDevice, 8, lowByte);
		if(retCode != SEAIO_ERROR_NONE)
		{
			return retCode;
		}

		retCode =	SeaIo_WriteReg(hDevice, 9, hiByte);
		if(retCode != SEAIO_ERROR_NONE)
		{
			return retCode;
		}

		/* Write any value to strobe the DAC */
		retCode =	SeaIo_WriteReg(hDevice, 10, 0);
		if(retCode != SEAIO_ERROR_NONE)
		{
			return retCode;
		}
	}
	else /* Channel is 1 */
	{
		retCode =	SeaIo_WriteReg(hDevice, 11, lowByte);
		if(retCode != SEAIO_ERROR_NONE)
		{
			return retCode;
		}

		retCode =	SeaIo_WriteReg(hDevice, 12, hiByte);
		if(retCode != SEAIO_ERROR_NONE)
		{
			return retCode;
		}

		/* Write any value to strobe the DAC */
		retCode =	SeaIo_WriteReg(hDevice, 13, 0);
		if(retCode != SEAIO_ERROR_NONE)
		{
			return retCode;
		}
	}

	/* Success - we are done with this channel */
	return SEAIO_ERROR_NONE;
}

int SeaIo_ReadReg(int hDevice, unsigned char bRegNumber,
		  unsigned char *pBuffer, unsigned long dwLength,
		  unsigned long *pDwRetLength)
{
	Seaio_Device			device						= {0};
	Seaio_ReadPortStruct	readPortStruct;
	unsigned char			readableAddressCount		= 0;
	int						success						= 0, 
							i							= 0;
	unsigned char			allReadableAddresses[32]	= {0};

	if(pBuffer == NULL || pDwRetLength == NULL)
		return SEAIO_ERROR_PARAM;

	if ( ioctl(hDevice, SEAIO_IOCGDESCRIPTOR, &device) < 0 ) {
		return Seaio_Error(errno);
	}

	for (i = 0; i < device.readable; i++) {
		allReadableAddresses[readableAddressCount] = device.input[i];
		readableAddressCount++;
	}
	for (i = 0; i < device.writable; i++) {
		allReadableAddresses[readableAddressCount] = device.output[i];
		readableAddressCount++;
	}
	
	if (readableAddressCount == 0)
		return SEAIO_ERROR_NOT_SUPPORTED;

	for (i = 0; i < readableAddressCount; i++) {
		if (allReadableAddresses[i] == bRegNumber) {
			readPortStruct.portOffset	= bRegNumber;
			readPortStruct.dataByte		= pBuffer;
			if(ioctl(hDevice, SEAIO_IOCREADPORT, &readPortStruct) < 0)
				return Seaio_Error(errno);
			*pDwRetLength = sizeof(unsigned char);
			success = 1;
			break;
		}
	}
	
	if (success != 1)
		return SEAIO_ERROR_PARAM;
	else
		return SEAIO_ERROR_NONE;
}


int SeaIo_RelayClose(int hDevice, unsigned char uRelayNum)
{
	return RelayToggle(hDevice, uRelayNum, CLOSE_RELAY);
}


int SeaIo_RelayOpen(int hDevice, unsigned char uRelayNum)
{
	return RelayToggle(hDevice, uRelayNum, OPEN_RELAY);
}


int SeaIo_GetData(int hDevice, unsigned char *pBuffer,
		  unsigned long dwLength, unsigned long *pDwRetLength)
{
	Seaio_Device device = {0};
	int i=0;
	unsigned char * dev_address_space = NULL;

	if ( ioctl(hDevice, SEAIO_IOCGDESCRIPTOR, &device) < 0 ) {
		return Seaio_Error(errno);
	}

	dev_address_space = (unsigned char *) malloc (device.io_addr_size);
	read(hDevice, dev_address_space, device.io_addr_size);
	for (i = 0; (i < MIN(device.io_addr_size, dwLength)) && (device.input[i] != -1); i++) {
		*(pBuffer + i) = dev_address_space[device.input[i]];
	}
	*pDwRetLength = i;
	free(dev_address_space);
	return SEAIO_ERROR_NONE;
}


int SeaIo_PutData(int hDevice, unsigned char *pBuffer)
{
	Seaio_Device device = {0};
	unsigned char port = 0, data = 0;
	unsigned char * dev_address_space = NULL;

	if ( ioctl(hDevice, SEAIO_IOCGDESCRIPTOR, &device) < 0 ) {
		return Seaio_Error(errno);
	}

	if (device.writable == 0)
		return SEAIO_ERROR_NOT_SUPPORTED;

	dev_address_space = (unsigned char *) malloc (device.io_addr_size);
	port = *pBuffer;
	data = *(pBuffer + 1);
	// set the port the way they want it
	if (port >= device.writable)
		return SEAIO_ERROR_PARAM;
	read(hDevice,dev_address_space,device.io_addr_size);
	dev_address_space[(device.output[port])] = data;
	write(hDevice,dev_address_space,device.io_addr_size);
	free(dev_address_space);
	return SEAIO_ERROR_NONE;
}


int SeaIo_GetAdapterInfo(int hDevice, pADAPTER_INFO pAdapter_Info)
{
	Seaio_Device device = {0};

	if ( ioctl(hDevice, SEAIO_IOCGDESCRIPTOR, &device) < 0 ) {
		return Seaio_Error(errno);
	}
	pAdapter_Info->wCardNumber = hextodec(device.model);
	pAdapter_Info->wBaseIO = device.io_base;
	pAdapter_Info->ucPortInCount = device.readable ;
	pAdapter_Info->ucPortOutCount = device.writable ;
	pAdapter_Info->ucPortCount = device.totalports ;
	pAdapter_Info->ucEnabled = 1;
	pAdapter_Info->wIRQ = device.irq_level;
	return SEAIO_ERROR_NONE;
}


int SeaIo_SetAdapterState(int hDevice, pADAPTER_STATE pAdapterState)
{
	Seaio_Device device = {0};
	c8255_Struct placeholder; 
	int i=0, numControlWords = 0;

	if ( ioctl(hDevice, SEAIO_IOCGDESCRIPTOR, &device) < 0 ) {
		return Seaio_Error(errno);
	}
	device.dwSampleInterval = pAdapterState->dwSampleInterval;
	for (i = 0; i < MAXCONTROL; i++)
		device.controlwords[i]=pAdapterState->ucModeCW[i];
	
	// Deal with configurable cards
	switch(device.model)
	{
		//
		// 8001 type
		//

		case 0x8001:
		case 0x8010:
			Setup8001(hDevice, &device);
			break;

		//
		// 8255 type
		//

		// 4 control words
		case 0x8009:
		case 0x8205:
			numControlWords ++;
			numControlWords ++;
		// 2 control words
		case 0x4030:
		case 4030:
		case 0x8005:
		case 0x8014:
		case 0x8203:
			numControlWords ++;
		// 1 control word
		case 0x8255:
		case 8255:
		case 0x8008:
			numControlWords ++;
			ClearSeaioDevice(&device);
			// now loop once for each control word
			for(i = 0; i < numControlWords; i ++)
			{
				Clear8255Struct(&placeholder);
				Setup8255(hDevice, &device, i, &placeholder);
			}
		break;
	}

	if ( ioctl(hDevice, SEAIO_IOCSDESCRIPTOR, &device) < 0 ) {
		return Seaio_Error(errno);
	}

	return SEAIO_ERROR_NONE;
}


int SeaIo_GetAdapterState(int hDevice, pADAPTER_STATE pAdapterState)
{
	Seaio_Device device = {0};
	int i=0;

	if ( ioctl(hDevice, SEAIO_IOCGDESCRIPTOR, &device) < 0 ) {
		return Seaio_Error(errno);
	}
	memset(pAdapterState,0,sizeof(ADAPTER_STATE));
	pAdapterState->dwSampleInterval = device.dwSampleInterval;
	for (i = 0; i < MAXCONTROL; i++) 
		pAdapterState->ucModeCW[i] = device.controlwords[i];
	return SEAIO_ERROR_NONE;
}


int Seaio_Error(int errornumber) 
{
	switch (errornumber) {
		case EBADF:
			return SEAIO_ERROR_PORT_NOT_OPEN;
			break;

		case ENOTTY:
			return SEAIO_ERROR_PORT_NOT_OPEN;
			break;

		case EFAULT:
			return SEAIO_ERROR_PARAM;
			break;

		case EINVAL:
			return SEAIO_ERROR_PARAM;
			break;

		default:
			return SEAIO_ERROR_GENERAL;
			break;
	}
}

int hextodec(unsigned short hexnum)
{
	int decimal = 0, i=0;
	for (i = 0; i < 4; i++)
		decimal += ((hexnum >> (i*4)) & 0x0f) * pow(10,i);
	return decimal;
}


int RelayToggle(int hDevice, unsigned char RelayNumber, unsigned char State)
{
	unsigned long	result, bufferSize = sizeof(unsigned char);
	unsigned char	dataByte;

	RelayNumber --;	// RelayNumber has historically been 1-based

	result = SeaIo_ReadReg(hDevice, RelayNumber / 8, &dataByte, bufferSize, &bufferSize);	
	if(result != SEAIO_ERROR_NONE)
		return result;

	if (State == CLOSE_RELAY) {
		dataByte = dataByte | (0x01 << (RelayNumber % 8));
	} else {
		dataByte = dataByte & ~(0x01 << (RelayNumber % 8));
	}
	
	return SeaIo_WriteReg(hDevice, RelayNumber / 8, dataByte);
}


void ClearSeaioDevice(pSeaio_Device device)
{
	int i = 0;

	for (i = 0; i < MAXINPUTS; i++) {
		device->input[i] = -1;
	}
	for (i = 0; i < MAXOUTPUTS; i++) {
		device->output[i] = -1;
	}
	device->writable = 0;
	device->readable = 0;
}

void Clear8255Struct(p8255_Struct my8255)
{
	int i = 0;
	
	// initialize my8255
	my8255->PortInCount = 0;
	my8255->PortOutCount = 0;
	for (i = 0; i < MAXINPUTS; i++) {
		my8255->input[i].portnum = -1;
		my8255->input[i].porttype = No_Port_Here;
	}
	for (i = 0; i < MAXOUTPUTS; i++) {
		my8255->output[i].portnum = -1;
		my8255->output[i].porttype = No_Port_Here;
	}
}

void Setup8001(int hDevice, pSeaio_Device device) 
{
	int i=0;
	Seaio_WritePortStruct	writePortStruct;
	
	// write the mode control word to the device
	writePortStruct.portOffset	= device->control[0];
	writePortStruct.dataByte	= device->controlwords[0];
	ioctl(hDevice, SEAIO_IOCWRITEPORT, &writePortStruct);

	//update port in and out counts
	ClearSeaioDevice(device);
	for (i = 0; i < 4; i++) {
		if ((device->controlwords[0] >> i)&0x01) {
			device->output[device->writable] = i;
			device->writable++;
		}else {
			device->input[device->readable] = i;
			device->readable++;
		}
	}
}

void Setup8255(int hDevice, pSeaio_Device device, int chipselect, p8255_Struct my8255)
{
	unsigned char ModeCW = device->controlwords[chipselect];
	int i = 0;
	Seaio_WritePortStruct	writePortStruct;
	
	// write the mode control word to the device
	writePortStruct.portOffset	= device->control[chipselect];
	writePortStruct.dataByte	= ModeCW;
	ioctl(hDevice, SEAIO_IOCWRITEPORT, &writePortStruct);
	
	// update port in and out counts
	
	// Since we know that the 8255 occupies 4 consecutive IO bytes, we can calculate the
	// offsets of ports A, B, and C from the 8255 Control Register offset
	if (ModeCW & 0x10) { // Port A is Input
		my8255->input[my8255->PortInCount].portnum = device->control[chipselect] - 3;
		my8255->input[my8255->PortInCount].porttype = Input;
		my8255->PortInCount++;
	} else { // Port A is Output
		my8255->output[my8255->PortOutCount].portnum = device->control[chipselect] - 3;
		my8255->output[my8255->PortOutCount].porttype = Output;
		my8255->PortOutCount++;
	}
	if (ModeCW & 0x02) { // Port B is Input
		my8255->input[my8255->PortInCount].portnum = device->control[chipselect] - 2;
		my8255->input[my8255->PortInCount].porttype = Input;
		my8255->PortInCount++;
	} else { // Port B is Output
		my8255->output[my8255->PortOutCount].portnum = device->control[chipselect] - 2;
		my8255->output[my8255->PortOutCount].porttype = Output;
		my8255->PortOutCount++;
	}
	switch (ModeCW & 0x09) 
	{
	case 0x09: // Both Hi and Low are Input
		my8255->input[my8255->PortInCount].portnum = device->control[chipselect] - 1;
		my8255->input[my8255->PortInCount].porttype = Input;
		my8255->PortInCount ++;
		break;

	case 0x00: // Both Hi and Low are Output
		my8255->output[my8255->PortOutCount].portnum = device->control[chipselect] - 1;
		my8255->output[my8255->PortOutCount].porttype = Output;
		my8255->PortOutCount ++;
		break;

	case 0x01: // Lo - In, Hi - Out
		// the 8009 uses 8255 but doesn't support split-c addressing, so anything except 0 here is input
		if(device->model == 8009 || device->model == 0x8009)
		{
			my8255->input[my8255->PortInCount].portnum = device->control[chipselect] - 1;
			my8255->input[my8255->PortInCount].porttype = LoIn_HiOut;
			my8255->PortInCount++;
			break;
		}

		my8255->input[my8255->PortInCount].portnum = device->control[chipselect] - 1;
		my8255->input[my8255->PortInCount].porttype = LoIn_HiOut;
		my8255->output[my8255->PortOutCount].portnum = device->control[chipselect] - 1;
		my8255->output[my8255->PortOutCount].porttype = LoIn_HiOut;
		my8255->PortInCount++;
		my8255->PortOutCount++;
		break;

	case 0x08: // Lo - Out, Hi - In
		// the 8009 uses 8255 but doesn't support split-c addressing, so anything except 0 here is input
		if(device->model == 8009 || device->model == 0x8009)
		{
			my8255->input[my8255->PortInCount].portnum = device->control[chipselect] - 1;
			my8255->input[my8255->PortInCount].porttype = LoIn_HiOut;
			my8255->PortInCount++;
			break;
		}

		my8255->input[my8255->PortInCount].portnum = device->control[chipselect] - 1;
		my8255->input[my8255->PortInCount].porttype = LoOut_HiIn;
		my8255->output[my8255->PortOutCount].portnum = device->control[chipselect] - 1;
		my8255->output[my8255->PortOutCount].porttype = LoOut_HiIn;
		my8255->PortInCount++;
		my8255->PortOutCount++;
		break;

	default: // I must have gone crazy if we get here
		break;
	}
	
	// update device with my8255 now
	for (i = 0; i < my8255->PortInCount; i++) {
		device->input[device->readable] = my8255->input[i].portnum;
		device->readable++;
	}
	for (i = 0; i < my8255->PortOutCount; i++) {
		device->output[device->writable] = my8255->output[i].portnum;
		device->writable++;
	}
}

int SeaIo_OpenDevice(unsigned char DeviceNumber, int * phDevice)
{
	char temp[13] = {0};
	snprintf(temp,13,"/dev/dio%d",DeviceNumber);
	*phDevice = open(temp,O_RDWR);
	if (*phDevice != -1) {
		return SEAIO_ERROR_NONE;
	} else {
		return errno;
	}
}

int SeaIo_CloseDevice(int hDevice)
{
	int status = SEAIO_ERROR_NONE;
	status = close(hDevice);
	if (status != 0) {
		if (errno == EBADF) {
			return SEAIO_ERROR_INVALID_HANDLE;
		} else {
			return SEAIO_ERROR_GENERAL;
		}
	} else {
		return SEAIO_ERROR_NONE;
	}
}

int SeaIo_StartInterruptSampling(int hDevice)
{
	Seaio_Device device = {0};
	int i=0;
	unsigned char * dev_address_space = NULL;

	if ( ioctl(hDevice, SEAIO_IOCGDESCRIPTOR, &device) < 0 ) {
		return Seaio_Error(errno);
	}

	dev_address_space = (unsigned char *) malloc (device.io_addr_size);

	read(hDevice,dev_address_space,device.io_addr_size);
	while (device.interrupt_control_port[i] != -1) {
		// 0x07 interrupts on rising edges,
		// 0x06 interrupts on falling edges,
		// 0x05 interrupts on high level,
		// 0x04 interrupts on low level
		dev_address_space[device.interrupt_control_port[0]] = 0x06;
		i++;
	}
	write(hDevice,dev_address_space,device.io_addr_size);
	free(dev_address_space);
	return 1;
}

int SeaIo_StopInterruptSampling(int hDevice)
{
	Seaio_Device device = {0};
	int i=0;
	unsigned char * dev_address_space = NULL;

	if ( ioctl(hDevice, SEAIO_IOCGDESCRIPTOR, &device) < 0 ) {
		return Seaio_Error(errno);
	}

	dev_address_space = (unsigned char *) malloc (device.io_addr_size);

	read(hDevice,dev_address_space,device.io_addr_size);
	while (device.interrupt_control_port[i] != -1) {
		// 0x00 always disables interrupts.
		dev_address_space[device.interrupt_control_port[i]] = 0x00;
		i++;
	}
	write(hDevice,dev_address_space,device.io_addr_size);
	free(dev_address_space);
	return 1;
}

int SeaIo_FlushInterruptBuffers(int hDevice)
{
	Seaio_Device device = {0};
	int i=0;
	unsigned char * dev_address_space = NULL;

	if ( ioctl(hDevice, SEAIO_IOCGDESCRIPTOR, &device) < 0 ) {
		return Seaio_Error(errno);
	}

	dev_address_space = (unsigned char *) malloc (device.io_addr_size);
	free(dev_address_space);
	i = 0;
	return 1;
}
